import inspect
import itertools
import types
from collections import deque

from pycsp3.classes.auxiliary.conditions import Condition
from pycsp3.classes.auxiliary.structures import Automaton
from pycsp3.classes.auxiliary.types import TypeOrderedOperator, TypeRank, TypeVar
from pycsp3.classes.entities import *
from pycsp3.classes.main.annotations import AnnotationDecision, AnnotationOutput, AnnotationVarHeuristic, AnnotationValHeuristic, \
    AnnotationFiltering, AnnotationPrepro, AnnotationSearch, AnnotationRestarts
from pycsp3.classes.main.constraints import *
from pycsp3.classes.main.domains import Domain
from pycsp3.classes.main.objectives import ObjectiveExpression, ObjectivePartial
from pycsp3.classes.main.variables import Variable, VariableInteger, VariableSymbolic, NotVariable, NegVariable
from pycsp3.tools.curser import OpOverrider, ListInt, ListVar
from pycsp3.tools.inspector import checkType, extract_declaration_for, comment_and_tags_of, comments_and_tags_of_parameters_of
from pycsp3.tools.utilities import flatten, is_1d_list, is_2d_list, is_matrix, is_square_matrix, to_alphabet_positions, transpose, is_containing

''' Global Variables '''

absPython = abs
maxPython = max
minPython = min


class Star(float):
    def __init__(self, val):
        super().__init__()

    def __str__(self):
        return "*"


ANY = Star("inf")

queue_in = deque()  # To keep partial constraints when using the IN operator


def protect():
    return OpOverrider.disable()


''' Checking Model Variants '''


def variant(name=None):
    pos = -1 if not options.variant else options.variant.find("-")  # position of dash ('-') in options.variant
    if not name:
        return options.variant[0:pos] if pos != -1 else options.variant
    return options.variant[0:pos] == name if pos != -1 else options.variant == name


def subvariant(name=None):
    pos = -1 if not options.variant else options.variant.find("-")  # position of dash ('-') in options.variant
    if not name:
        return None if pos == -1 else options.variant[pos + 1:]
    return pos != -1 and options.variant[pos + 1:] == name


''' Declaring stand-alone variables and arrays '''


def Var(term=None, *others, dom=None):
    assert (term is None) != (dom is None)
    if term is not None:
        dom = flatten(term, others)
    if not isinstance(dom, Domain):
        dom = Domain(dom)
    name = extract_declaration_for("Var")
    comment, tags = comment_and_tags_of(function_name="Var")

    assert isinstance(comment, (str, type(None))), \
        "A comment must be a string (or None). Usually, they are given on plain lines preceding the declaration"
    assert name not in Variable.name2obj, \
        "The identifier " + name + " is used twice. This is not possible"
    assert dom.get_type() in {TypeVar.INTEGER, TypeVar.SYMBOLIC}, \
        "Currently, only integer and symbolic variables are supported. Problem with domain " + str(dom)
    assert str(name) not in Variable.name2obj, "The identifier " + name + " is used twice. This is not possible"

    var_object = VariableInteger(name, dom) if dom.get_type() == TypeVar.INTEGER else VariableSymbolic(name, dom)
    Variable.name2obj[name] = var_object
    EVar(var_object, comment, tags)  # object wrapping the variable x
    return var_object


def VarArray(*, size, dom, when=None, comment=None):
    size = [size] if isinstance(size, int) else size
    assert all(dimension != 0 for dimension in size), "No dimension must not be equal to 0"
    checkType(size, [int])
    checkType(dom, (int, range, Domain, [int, range, str, Domain, type(None)], type(lambda: 0)))
    name = extract_declaration_for("VarArray")
    if comment is None and not isinstance(name, list):
        comment, tags = comment_and_tags_of(function_name="VarArray")

    assert isinstance(comment, (str, type(None))), \
        "A comment must be a string (or None). Usually, they are given on plain lines preceding the declaration"
    assert not isinstance(dom, type(lambda: 0)) or len(size) == len(inspect.signature(dom).parameters), \
        "The number of arguments of the lambda must be equal to the number of dimensions of the multidimensional array"
    assert isinstance(comment, (str, type(None))), \
        "A comment must be a string (or None). Usually, they are given on plain lines preceding the declaration"
    assert str(name) not in Variable.name2obj, "The identifier " + name + " is used twice. This is not possible"

    var_objects = Variable.build_variables_array(name, size, dom, when)
    if isinstance(name, list):
        for variable in var_objects:
            EVar(variable, None, None)  # object wrapping the variables
        return tuple(var_objects)
    else:
        EVarArray(var_objects, name, comment, tags)  # object wrapping the array of variables
        return ListVar(var_objects)


''' Posting constraints (through satisfy()) '''


def _bool_interpretation_for_in(left_operand, right_operand, bool_value):
    assert type(bool_value) is bool
    if isinstance(left_operand, Variable) and isinstance(right_operand, (set, frozenset, range)):
        # It is a unary constraint of the form x in/not in set/range
        ctr = Intension(isin(left_operand, right_operand) if bool_value else notin(left_operand, right_operand))
    elif isinstance(left_operand, PartialConstraint):  # It is a partial form of constraint (sum, count, maximum, ...)
        ctr = ECtr(left_operand.constraint.replace_condition(TypeConditionOperator.IN if bool_value else TypeConditionOperator.NOTIN, right_operand))
    elif isinstance(right_operand, Automaton):  #  It is a regular constraint
        ctr = Regular(scope=left_operand, automaton=right_operand)
    else:  #  It is a table constraint
        if not hasattr(left_operand, '__iter__'):
            left_operand = [left_operand]
        ctr = Extension(scope=list(left_operand), table=right_operand, positive=bool_value)
    return ctr


def _complete_partial_forms_of_constraints(entities):
    for i, c in enumerate(entities):
        if isinstance(c, bool):
            assert len(queue_in) > 0, "A boolean that does not represent a constraint is in the list of constraints in satisfy(): " + str(entities)
            right_operand, left_operand = queue_in.popleft()
            entities[i] = _bool_interpretation_for_in(left_operand, right_operand, c)
        elif isinstance(c, ESlide):
            for ent in c.entities:
                _complete_partial_forms_of_constraints(ent.entities)
    return entities


def _wrap_intension_constraints(entities):
    for i, c in enumerate(entities):
        if isinstance(c, Node):
            entities[i] = Intension(c)  # the node is wrapped by a Constraint object (Intension)
    return entities


def _group(*args):
    entities = _wrap_intension_constraints(_complete_partial_forms_of_constraints(flatten(*args)))
    checkType(entities, allowedTypes=[ECtr, ECtrs])
    return EToGather(entities)


def _reorder(entities):
    reordered_entities = []
    g = []
    for c in entities:
        if isinstance(c, ECtr):
            g.append(c)
        else:
            if len(g) != 0:
                reordered_entities.append(_group(g))
            g.clear()
            reordered_entities.append(c)
    if len(g) != 0:
        reordered_entities.append(_group(g))
    return reordered_entities


def _block(*args):
    entities = _wrap_intension_constraints(_complete_partial_forms_of_constraints(flatten(*args)))
    checkType(entities, allowedTypes=[ECtr, ECtrs])
    entities = _reorder(entities)
    return EBlock(entities)


def IfThenElse(*args):
    entities = _wrap_intension_constraints(_complete_partial_forms_of_constraints(flatten(*args)))
    checkType(entities, allowedTypes=[ECtr])
    assert len(entities) == 3, "Error: three components must be specified in ifThenElse"
    return EIfThenElse(entities)


def Slide(*args):
    entities = _wrap_intension_constraints(flatten(*args))  # we cannot directly complete partial forms (because it may be called by the user?)
    checkType(entities, allowedTypes=[ECtr, bool])
    return ESlide([EToGather(entities)])


def satisfy(*args):
    global no_parameter_satisfy
    global nb_parameter_satisfy
    no_parameter_satisfy = 0
    nb_parameter_satisfy = len(args)
    comments1, comments2, tags1, tags2 = comments_and_tags_of_parameters_of(function_name="satisfy", args=args)
    t = []
    for i, arg in enumerate(args):
        no_parameter_satisfy = i
        assert isinstance(arg, (ECtr, ESlide, Node, bool, list, tuple, type(None), types.GeneratorType)), \
            "Authorized satisfy()'s type parameter are ECtr, ESlide, Node, bool, list, tuple, None or a generator. Not " + str(type(arg))
        if arg is None:
            continue
        arg = list(arg) if isinstance(arg, types.GeneratorType) else arg
        comment_at_2 = any(comment != '' for comment in comments2[i])
        if isinstance(arg, (ECtr, ESlide)):  # case: simple constraint or slide
            to_post = _complete_partial_forms_of_constraints([arg])[0]
        elif isinstance(arg, Node):  # a predicate to be wrapped by a constraint (intension)
            to_post = Intension(arg)
        elif isinstance(arg, bool):  # a Boolean representing the case of a partial constraint or a node with operator in {IN, NOT IN}
            assert queue_in, "A boolean that do not represents a constraint is in the list of constraints in satisfy(): " \
                             + str(args) + ".\nA constraint is certainly wrong"
            other, partial = queue_in.popleft()
            to_post = _bool_interpretation_for_in(partial, other, arg)
        elif any(isinstance(ele, ESlide) for ele in arg):  #  Case: Slide
            to_post = _block(arg)
        elif comment_at_2:  #  Case: block
            for j, ele in enumerate(arg):
                if isinstance(arg[j], (ECtr, ESlide)):
                    arg[j].note(comments2[i][j]).tag(tags2[i][j])
                elif comments2[i][j] or tags2[i][j]:
                    arg[j] = _group(arg[j]).note(comments2[i][j]).tag(tags2[i][j])
            to_post = _block(arg)
        else:  #  Group
            to_post = _group(arg)
        if to_post:
            t.append(to_post.note(comments1[i]).tag(tags1[i]))
    return EToSatisfy(t)


''' Generic Constraints (intension, extension) '''

checked_tables = set()


def Extension(*, scope, table, positive=True):
    scope = flatten(scope)
    checkType(scope, [Variable])
    checkType(table, [str, int, float])
    checkType(positive, bool)
    assert isinstance(table, list) and len(table) > 0, "A table must be a non-empty list of tuples"
    assert isinstance(table[0], (tuple, int)), "Table in extension must be a list of tuples or a list of int"
    if not isinstance(table[0], int):
        assert len(scope) == len(table[0]), "Scope and tuples of the table must be of the same length"
    # TODO: this ckecking don't pass on Waterbucket.py, but the xml file is the same that the java version !
    # if options.checker:
    #    if id(table) not in checked_tables:
    #        for t in table:
    #            for i, v in enumerate(t):
    #                if v not in variables[i].dom:
    #                    raise ValueError(
    #                        "Problem: Constraint extension : a value of table is not represented in a domain of a variable : " + str(domainElement))
    #        checked_tables.add(id(table))
    return ECtr(ConstraintExtension(scope, table, positive))


''' Constraints in intension '''


def Intension(node):
    checkType(node, Node)
    ctr = ECtr(ConstraintIntension(node))
    if ctr.blank_basic_attributes():
        ctr.copy_basic_attributes_of(node)
    node.mark_as_used()
    return ctr


def abs(*args):
    return Node.build(TypeNode.ABS, *args) if len(args) == 1 and isinstance(args[0], (Node, Variable)) else absPython(*args)


def neg(*args):
    return Node.build(TypeNode.NEG, *args)


def sqr(*args):
    return Node.build(TypeNode.SQR, *args)


def add(*args):
    return Node.build(TypeNode.ADD, *args)


def sub(*args):
    return Node.build(TypeNode.SUB, *args)


def mul(*args):
    return Node.build(TypeNode.MUL, *args)


def div(*args):
    return Node.build(TypeNode.DIV, *args)


def mod(*args):
    return Node.build(TypeNode.MOD, *args)


def power(*args):
    return Node.build(TypeNode.POW, *args)


def min(*args):
    return Node.build(TypeNode.MIN, *args) if len(args) > 0 and any(isinstance(a, (Node, Variable)) for a in args) else minPython(*args)


def max(*args):
    return Node.build(TypeNode.MAX, *args) if len(args) > 0 and any(isinstance(a, (Node, Variable)) for a in args) else maxPython(*args)


def dist(*args):
    return Node.build(TypeNode.DIST, *args)


def lt(*args):
    return Node.build(TypeNode.LT, *args)


def le(*args):
    return Node.build(TypeNode.LE, *args)


def ge(*args):
    return Node.build(TypeNode.GE, *args)


def gt(*args):
    return Node.build(TypeNode.GT, *args)


def ne(*args):
    return Node.build(TypeNode.NE, *args)


def eq(*args):
    return Node.build(TypeNode.EQ, *args)


def isin(*args):
    return Node.build(TypeNode.IN, *args)


def notin(*args):
    return Node.build(TypeNode.NOTIN, *args)


def lnot(*args):
    return Node.build(TypeNode.NOT, *args)


def land(*args):
    return Node.build(TypeNode.AND, *args) if len(args) > 1 else args[0]


def lor(*args):
    return Node.build(TypeNode.OR, *args) if len(args) > 1 else args[0]


def xor(*args):
    return Node.build(TypeNode.XOR, *args) if len(args) > 1 else args[0]


def iff(*args):
    return Node.build(TypeNode.IFF, *args)


def imply(*args):
    return Node.build(TypeNode.IMP, *args)


def ift(*args):
    return Node.build(TypeNode.IF, *args)


def cunjunction(*args):
    if len(args) == 1:
        if isinstance(args[0], list):
            args = tuple(args[0])
        if isinstance(args[0], types.GeneratorType):
            args = tuple(list(args[0]))
    return land(*args)


def disjunction(*args):
    if len(args) == 1:
        if isinstance(args[0], list):
            args = tuple(args[0])
        if isinstance(args[0], types.GeneratorType):
            args = tuple(list(args[0]))
    return lor(*args)


def knight_attack(x, y, order):
    d1, d2 = dist(x // order, y // order), dist(x % order, y % order)  # distances between rows and columns
    return (d1 == 1) & (d2 == 2) | (d1 == 2) & (d2 == 1)


def queen_attack(x, y, order):
    d1, d2 = dist(x // order, y // order), dist(x % order, y % order)  # distances between rows and columns
    return (x != y) & ((x % order == y % order) | (x // order == y // order) | (d1 == d2))


''' Language-based Constraints '''


def Regular(*, scope, automaton):
    scope = flatten(scope)
    checkType(scope, [Variable])
    checkType(automaton, Automaton)
    return ECtr(ConstraintRegular(scope, automaton))


def Mdd(*, scope, transitions):
    scope = flatten(scope)
    checkType(scope, [Variable])
    checkType(transitions, list)
    return ECtr(ConstraintMdd(scope, transitions))


''' Comparison-based Constraints '''


def AllDifferent(term, *others, excepting=None, matrix=None):
    terms = flatten(term, others)
    if matrix is not None:
        assert excepting is None, "excepting values are currently not supported for AllDifferentMatrix"
        return AllDifferentMatrix(terms)
    excepting = list(excepting) if isinstance(excepting, (tuple, set)) else [excepting] if isinstance(excepting, int) else excepting
    checkType(terms, ([Variable], [Node]))
    checkType(excepting, ([int], type(None)))
    return ECtr(ConstraintAllDifferent(terms, excepting))


def AllDifferentList(lists, *others, excepting=None):
    if len(others) > 0:
        lists = list((lists,) + others)
    lists = [flatten(l) for l in lists]
    for l in lists:
        checkType(l, [Variable])
    excepting = list(excepting) if isinstance(excepting, (tuple, set)) else excepting
    checkType(excepting, ([int], type(None)))
    assert all(len(l) == len(lists[0]) for l in lists) and (excepting is None or len(excepting) == len(list[0]))
    return ECtr(ConstraintAllDifferentList(lists, excepting))


def AllDifferentMatrix(matrix):
    matrix = [flatten(row) for row in matrix]
    assert all(len(row) == len(matrix[0]) for row in matrix), "The matrix id badly formed"
    for l in matrix:
        checkType(l, [Variable])
    return ECtr(ConstraintAllDifferentMatrix(matrix))


def AllEqual(term, *others):
    terms = flatten(term, others)
    checkType(terms, ([Variable], [Node]))
    return ECtr(ConstraintAllEqual(terms))


def _ordered(term, *others, operator, lengths=None):
    terms = flatten(term, others)
    checkType(terms, [Variable])
    checkType(operator, TypeOrderedOperator)
    checkType(lengths, ([int, Variable], type(None)))
    if lengths is not None:
        if len(terms) == len(lengths):
            lengths = lengths[:-1]  # we assume that the last value is useless
        else:
            assert len(terms) == len(lengths) + 1
    return ECtr(ConstraintOrdered(terms, operator, lengths))


def Increasing(term, *others, strict=False, lengths=None):
    return _ordered(term, *others, operator=TypeOrderedOperator.INCREASING if not strict else TypeOrderedOperator.STRICTLY_INCREASING, lengths=lengths)


def Decreasing(term, *others, strict=False, lengths=None):
    return _ordered(term, *others, operator=TypeOrderedOperator.DECREASING if not strict else TypeOrderedOperator.STRICTLY_DECREASING, lengths=lengths)


def _lex(term, *others, operator, matrix=False):
    if len(others) == 0:
        assert is_matrix(term, Variable)
        lists = [flatten(l) for l in term]
    else:
        assert is_1d_list(term, Variable) and all(is_1d_list(l, Variable) for l in others)
        lists = [flatten(term)] + [flatten(l) for l in others]
    assert is_matrix(lists, Variable)  # new check because some null cells (variables) may have been discarded
    checkType(lists, ([Variable]))
    assert all(len(l) == len(lists[0]) for l in lists)
    checkType(operator, TypeOrderedOperator)
    if matrix:
        return ECtr(ConstraintLexMatrix(lists, operator))
    return ECtr(ConstraintLex(lists, operator))


def LexIncreasing(term, *others, strict=False, matrix=False):
    return _lex(term, *others, operator=TypeOrderedOperator.INCREASING if not strict else TypeOrderedOperator.STRICTLY_INCREASING, matrix=matrix)


def LexDecreasing(term, *others, strict=False, matrix=False):
    return _lex(term, *others, operator=TypeOrderedOperator.DECREASING if not strict else TypeOrderedOperator.STRICTLY_DECREASING, matrix=matrix)


''' Method for handling complete/partial constraints '''


def _wrapping_by_complete_or_partial_constraint(ctr):
    condition = ctr.arguments[TypeCtrArg.CONDITION].content
    return ECtr(ctr) if condition is not None else PartialConstraint(ctr)


''' Counting and Summing Constraints '''


def _manage_coeffs(terms, coeffs):
    if coeffs:
        OpOverrider.disable()
        if len(coeffs) == 1 and isinstance(coeffs[0], (tuple, set, range)):
            coeffs = list(coeffs[0])
        elif isinstance(coeffs, (tuple, set, range)):
            coeffs = list(coeffs)
        elif isinstance(coeffs, (int, Variable)):
            coeffs = [coeffs]
        assert len(terms) == len(coeffs), \
            "Lists (variables and coefficients) should have the same length." \
            + " Here, we have " + str(len(terms)) + "!=" + str(len(coeffs))
        # if 0 in coeffs:
        #    terms = [terms[i] for i in range(len(terms)) if coeffs[i] != 0]
        #    coeffs = [c for c in coeffs if c != 0]
        if all(c == 1 for c in coeffs):
            coeffs = None
        checkType(coeffs, allowedTypes=([Variable, int], type(None)))
        OpOverrider.enable()
    return terms, coeffs


def _get_terms_coeffs(terms):
    if len(terms) == 1 and isinstance(terms[0], ScalarProduct):
        return flatten(terms[0].variables), flatten(terms[0].coeffs)
    if all(isinstance(x, Variable) for x in terms):
        return terms, None
    t1, t2 = [], []
    for tree in terms:
        if isinstance(tree, Variable):
            t1.append(tree)
            t2.append(1)
        else:
            assert isinstance(tree, (Node, NegVariable))
            pair = (tree.variable, -1) if isinstance(tree, NegVariable) else tree.tree_val_if_binary_type(TypeNode.MUL)
            if pair is None:
                break
            t1.append(pair[0])
            t2.append(pair[1])
    if len(t1) == len(terms):
        for tree in terms:
            if isinstance(tree, Node):
                tree.mark_as_used()
        return t1, t2
    return terms, None


def Sum(term, *others, condition=None):
    """Builds a constraint http://xcsp.org/specifications/sum from the specified arguments
    """
    term = list(term) if isinstance(term, types.GeneratorType) else term
    checkType(term, ([Variable], [Node], Variable, Node, ScalarProduct))
    for other in others:
        checkType(other, ([Variable], [Node], Variable, Node, ScalarProduct))
    terms = list(term) if isinstance(term, types.GeneratorType) else flatten(term, others)

    terms, coeffs = _get_terms_coeffs(terms)
    terms, coeffs = _manage_coeffs(terms, coeffs)
    # TODO control here some assumptions
    return _wrapping_by_complete_or_partial_constraint(ConstraintSum(terms, coeffs, Condition.build_condition(condition)))


def Count(term, *others, value=None, values=None, condition=None):
    terms = flatten(term, others)
    assert len(terms) > 0, "A count with an empty scope"
    checkType(terms, ([Variable], [Node]))
    if value is None and values is None:
        value = 1
    assert value is not None or values is not None, str(value) + " " + str(values)
    values = list(values) if isinstance(values, (tuple, set)) else [value] if isinstance(value, (int, Variable)) else values
    checkType(values, allowedTypes=([int], [Variable]))
    return _wrapping_by_complete_or_partial_constraint(ConstraintCount(terms, values, Condition.build_condition(condition)))


# def AtLeastOne(term, *others, value=None, values=None):
#     return Count(term, others, value=value, values=values, condition=(TypeConditionOperator.GE, 1))
#
#
# def AtMostOne(term, *others, value=None, values=None):
#     return Count(term, others, value=value, values=values, condition=(TypeConditionOperator.LE, 1))
#
#
# def ExactlyOne(term, *others, value=None, values=None):
#     return Count(term, others, value=value, values=values, condition=(TypeConditionOperator.EQ, 1))


def Nvalues(term, *others, condition=None, excepting=None):
    terms = flatten(term, others)
    checkType(terms, [Variable])
    if excepting is not None:
        excepting = flatten(excepting)
    checkType(excepting, ([int], type(None)))
    return _wrapping_by_complete_or_partial_constraint(ConstraintNValues(terms, Condition.build_condition(condition), excepting))


def NotAllEqual(term, *others):
    return Nvalues(term, others, condition=(TypeConditionOperator.GT, 1))


def Cardinality(term, *others, occurrences, closed=None):
    terms = flatten(term, others)
    checkType(terms, ([Variable]))
    assert isinstance(occurrences, dict)
    values = list(occurrences.keys())
    assert all(isinstance(value, (int, Variable)) for value in values)
    occurs = list(occurrences.values())
    checkType(closed, (bool, type(None)))
    for i, occ in enumerate(occurs):
        if isinstance(occ, range):
            occurs[i] = str(min(occ)) + ".." + str(max(occ))
        if isinstance(occ, list):
            flat = flatten(occ)
            if all([isinstance(e, int) for e in flat]) and flat == list(range(min(flat), max(flat) + 1)):
                flat = str(min(flat)) + ".." + str(max(flat))
            occurs[i] = flat
    # if len(occurs) == 1 and len(values) > 1:
    #     occurs = [occurs[0]] * len(values)
    return ECtr(ConstraintCardinality(terms, values, occurs, closed))


''' Connection Constraints '''


def Maximum(term, *others, index=None, startIndex=None, typeRank=None, condition=None):
    terms = flatten(term, others)
    checkType(terms, [Variable])
    checkType(index, (Variable, type(None)))
    checkType(startIndex, (int, type(None)))
    checkType(typeRank, (TypeRank, type(None)))
    if index is None:
        assert startIndex is None, "The argument startIndex must be used with index as explicit argument."
        assert typeRank is None, "The argument typeRank must be used with index as explicit argument."
    return _wrapping_by_complete_or_partial_constraint(ConstraintMaximum(terms, Condition.build_condition(condition), index, startIndex, typeRank))


def Minimum(term, *others, index=None, startIndex=None, typeRank=None, condition=None):
    terms = flatten(term, others)
    checkType(terms, [Variable])
    checkType(index, (Variable, type(None)))
    checkType(startIndex, (int, type(None)))
    checkType(typeRank, (TypeRank, type(None)))
    if index is None:
        assert startIndex is None, "The argument startIndex must be used with index as explicit argument."
        assert typeRank is None, "The argument typeRank must be used with index as explicit argument."

    return _wrapping_by_complete_or_partial_constraint(ConstraintMinimum(terms, Condition.build_condition(condition), index, startIndex, typeRank))


def Element(*, vector, index=None, value, startIndex=None, rank=TypeRank.ANY):
    vector = flatten(vector)
    checkType(vector, ([Variable], [int]))
    checkType(index, (Variable, type(None)))
    checkType(value, (Variable, int))
    checkType(startIndex, (int, type(None)))
    checkType(rank, TypeRank)
    if index is None:
        assert startIndex is None, "the startIndex argument must be used with index as explicit argument."
        assert rank is None, "the rank argument must be used with index as explicit argument."

    return ECtr(ConstraintElement(vector, value, index, startIndex if startIndex != 0 else None, rank if rank != TypeRank.ANY else None))


def Channel(list1, list2=None, *, startIndex1=None, startIndex2=None):
    list1 = flatten(list1)
    checkType(list1, [Variable])
    if list2 is not None: list2 = flatten(list2)
    checkType(list2, ([Variable], type(None)))
    checkType(startIndex1, (int, type(None)))
    checkType(startIndex2, (int, type(None)))
    if startIndex2 is not None:
        assert list2 is None, "the startIndex2 argument must be used with list2 as explicit argument."
    return ECtr(ConstraintChannel(list1, startIndex1, list2, startIndex2))


''' Packing and Scheduling Constraints '''


def NoOverlap(*, origins, lengths, zeroIgnored=None):
    # Warning, no flatten for this constraint, it is tuples
    # origins, lengths = flattens(origins, lengths)
    # if isinstance(origins,Variable):origins=[origins]
    # if isinstance(lengths, (Variable,int)): lengths = [lengths]
    checkType(origins, allowedTypes=([Variable]))
    checkType(lengths, allowedTypes=([Variable], [int]))
    return ECtr(ConstraintNoOverlap(origins, lengths, zeroIgnored))


def Cumulative(*, origins, lengths, heights, ends=None, condition=None):
    origins = flatten(origins)
    checkType(origins, allowedTypes=([Variable]))
    lengths = flatten(lengths)
    checkType(lengths, allowedTypes=([Variable], [int]))
    heights = flatten(heights)
    checkType(heights, allowedTypes=([Variable], [int]))
    if ends is not None: ends = flatten(ends)
    checkType(ends, allowedTypes=([Variable], type(None)))
    return _wrapping_by_complete_or_partial_constraint(ConstraintCumulative(origins, lengths, heights, Condition.build_condition(condition), ends))


''' Constraints on Graphs'''


def Circuit(term, *others, startIndex=None, size=None):
    terms = flatten(term, others)
    checkType(terms, [Variable])
    checkType(startIndex, (int, type(None)))
    checkType(size, (int, type(None)))
    return ECtr(ConstraintCircuit(terms, startIndex, size))


''' Elementary Constraints '''


def Clause(term, *others, phases=None):
    literals = flatten(term, others)
    phases = [False] * len(literals) if phases is None else flatten(phases)
    assert len(literals) == len(phases)
    checkType(literals, ([Variable, NotVariable]))
    checkType(phases, [bool])
    for i, literal in enumerate(literals):
        if isinstance(literal, NotVariable):
            literals[i] = literal.variable
            phases[i] = True
    return ECtr(ConstraintClause(literals, phases))


def Instantiation(*, variables, values):
    variables = flatten(variables)
    values = flatten(values) if not isinstance(values, range) else list(values)
    checkType(variables, [Variable])
    checkType(values, (int, [int]))
    if len(variables) == 0:
        return ECtr(None)
    if len(values) == 1 and len(variables) > 1:
        values = [values[0]] * len(variables)
    return ECtr(ConstraintInstantiation(variables, values))


''' Objectives '''


def _optimize(term, minimization):
    if isinstance(term, ScalarProduct):
        term = Sum(term)  # to have a PartialConstraint
    checkType(term, (Variable, Node, PartialConstraint)), "Did you forget to use Sum, e.g., as in Sum(x[i]*3 for i in range(10))"
    comment, _, tag, _ = comments_and_tags_of_parameters_of(function_name="minimize" if minimization else "maximize", args=[term])
    type = TypeCtr.MINIMIZE if minimization else TypeCtr.MAXIMIZE
    if isinstance(term, (Variable, Node)):
        if isinstance(term, Node):
            term.mark_as_used()
        return EObjective(ObjectiveExpression(type, term)).note(comment[0]).tag(tag[0])  # TODO why only one tag?
    else:
        return EObjective(ObjectivePartial(type, term)).note(comment[0]).tag(tag[0])


def minimize(term):
    return _optimize(term, True)


def maximize(term):
    return _optimize(term, False)


''' Annotations '''


def annotate(*, decision=None, output=None, varHeuristic=None, valHeuristic=None, filtering=None, prepro=None, search=None, restarts=None):
    def _new_annotation(obj):
        assert type(obj) not in AnnEntities.items_types, "This annotation can be specified only one time"
        return EAnnotation(obj)

    annotations = []
    if decision:
        annotations.append(_new_annotation(AnnotationDecision(decision)))
    if output:
        annotations.append(_new_annotation(AnnotationOutput(output)))
    if varHeuristic:
        annotations.append(_new_annotation(AnnotationVarHeuristic(varHeuristic)))
    if valHeuristic:
        annotations.append(_new_annotation(AnnotationValHeuristic(valHeuristic)))
    if filtering:
        annotations.append(_new_annotation(AnnotationFiltering(filtering)))
    if prepro:
        annotations.append(_new_annotation(AnnotationPrepro(prepro)))
    if search:
        annotations.append(_new_annotation(AnnotationSearch(search)))
    if restarts:
        annotations.append(_new_annotation(AnnotationRestarts(restarts)))
    return annotations


''' Helpers '''


def column(m, j):
    assert is_2d_list(m), "column() can only be called on 2-dimensional lists"
    assert all(len(row) > j for row in m), "one row has not at least j+1 elements"
    return ListVar(row[j] for row in m)


def columns(m):
    assert is_matrix(m), "columns() can only be called on matrices"
    return ListVar(column(m, j) for j in range(len(m[0])))


def diagonal_down(m, i=-1, j=-1, check=True):
    if check is True:
        assert is_square_matrix(m), "The specified first parameter must be a square matrix."
    if i == -1 and j == -1:
        return diagonal_down(m, 0, 0, False)
    if j == -1:
        return ListVar(m[k][len(m) - (i - k) if k < i else k - i] for k in range(len(m)))
    return ListVar(m[i + k][j + k] for k in range(len(m) - max(i, j)))


def diagonals_down(m, *, broken=False):
    assert is_square_matrix(m), "The specified first parameter must be a square matrix."
    if broken:
        return ListVar(diagonal_down(m, i, -1, False) for i in range(len(m)))
    return ListVar(diagonal_down(m, i, 0, False) for i in reversed(range(len(m) - 1))) + ListVar(
        diagonal_down(m, 0, j, False) for j in range(1, len(m) - 1))


def diagonal_up(m, i=-1, j=-1, check=True):
    if check is True:
        assert is_square_matrix(m), "The specified first parameter must be a square matrix."
    if i == -1 and j == -1:
        return diagonal_up(m, len(m) - 1, 0, False)
    if j == -1:
        return ListVar(m[k][len(m) - i - k - 1 if k < len(m) - i else 2 * len(m) - i - k - 1] for k in range(len(m)))
    return ListVar(m[i - k][j + k] for k in range(min(i + 1, len(m) - j)))


def diagonals_up(m, *, broken=False):
    assert is_square_matrix(m), "The specified first parameter must be a square matrix."
    if broken:
        return ListVar(diagonal_up(m, i, -1, False) for i in range(len(m)))
    return ListVar(diagonal_up(m, i, 0, False) for i in range(1, len(m))) + ListVar(
        diagonal_up(m, len(m) - 1, j, False) for j in range(1, len(m) - 1))


def different_values(*args):
    assert all(isinstance(arg, int) for arg in args)
    return all(a != b for (a, b) in itertools.combinations(args, 2))


def to_ordinary_table(table, domain_sizes):
    otable = []
    for t in table:
        if ANY in t:
            for otuple in itertools.product(*(range(v, v + 1) if v != ANY else range(domain_sizes[i]) for i, v in enumerate(t))):
                otable.append(otuple)
        else:
            otable.append(t)
    return otable


def cp_array(l):
    if isinstance(l, (tuple, set, types.GeneratorType)):
        l = list(l)
    assert isinstance(l, list) and len(l) > 0
    if isinstance(l[0], (list, types.GeneratorType)):
        assert all(isinstance(t, (list, types.GeneratorType)) for t in l)
        res = [cp_array(t) for t in l]
        return ListInt(res) if isinstance(res[0], ListInt) else ListVar(res)
    if all(isinstance(v, int) for v in l):
        return ListInt(l)
    elif all(isinstance(v, Variable) for v in l):
        return ListVar(l)
    else:
        raise NotImplemented


# def indexing(l):
#    assert isinstance(l, list)
#    return [(i, *v) if isinstance(v, (tuple, list)) else (i, v) for i, v in enumerate(l)]

import pycsp3.tools.curser


def _pycharm_security():
    _ = pycsp3.tools.curser
    _ = to_alphabet_positions
    _ = transpose
    _ = is_containing
    _ = CtrEntities
